#*****************************************************************************
# Script handles all tasks with Sun audio file (writing, reading, tags...) 
#
# Author: Falk Richter, falk.richter@yandex.com, 2021 
# $Id: AuFile.tcl,v 1.3 2021/12/08 20:21:42 falk Exp falk $
#*****************************************************************************

###########################################################
# Generate a new namespace
###########################################################
namespace eval ::AuFile {
  variable Local; 
  
  namespace export RipTracks AddTags ReadTags AuToRaw GetErrorMsg \
  GenreCat Comment

  # the genre catalogue for ID3v1
  set GenreCat [list Blues "Classic Rock" Country Dance \
  Disco Funk Grunge "Hip-Hop" Jazz Metal "New Age" Oldies Other Pop \
  R&B Rap Reggae Rock Techno Industrial Alternative Ska "Death Metal" \
  Pranks Soundtrack "Euro-Techno" Ambient "Trip-Hop" Vocal "Jazz+Funk" \
  Fusion Trance Classical Instrumental Acid House Game "Sound Clip" \
  Gospel Noise AlternRock Bass Soul Punk Space Meditative \
  "Instrumental Pop" "Instrumental Rock" Ethnic Gothic Darkwave \
  "Techno-Industrial" Electronic "Pop-Folk" Eurodance Dream \
  "Southern Rock" Comedy Cult Gangsta "Top 40" "Christian Rap" \
  "Pop/Funk" Jungle "Native American" Cabaret "New Wave" Psychadelic \
  Rave Showtunes Trailer "Lo-Fi" Tribal "Acid Punk" "Acid Jazz" Polka \
  Retro Musical "Rock & Roll" "Hard Rock" Folk "Folk-Rock" \
  "National Folk" Swing "Fast Fusion" Bebob Latin Revival Celtic \
  Bluegrass Avantgarde "Gothic Rock" "Progressive Rock" \
  "Psychedelic Rock" "Symphonic Rock" "Slow Rock" "Big Band" Chorus \
  "Easy Listening" Acoustic Humour Speech Chanson Opera "Chamber Music" \
  Sonata Symphony "Booty Brass" Primus "Porn Groove" Satire "Slow Jam" \
  Club Tango Samba Folklore Ballad "Power Ballad" "Rhytmic Soul" \
  Freestyle Duet "Punk Rock" "Drum Solo" "A Capela" "Euro-House" \
  "Dance Hall"]

  #  we calculate the length of the genre catalogue only once
  set Local(GenreCatLen) [llength $GenreCat] 

  # the comment is also used in the tagger tool
  set Comment "Created by AuTags"

  # preset the OS type (to have the variable in any case)
  set Local(OS) ""
}


###########################################################
# Set the OS type (needed for cdparanoia)
########################################################### 
proc ::AuFile::SetOsType { OsType } {
  variable Local

  set Local(OS) $OsType
}


###########################################################
# Open an au-file for reading or writing
###########################################################
proc ::AuFile::OpenFile { FileName {Type "Read"} } {
  variable Local 
 
  # preset the local file id 
  set Local(FileId) "" 

  # we set the access type
  switch $Type {
	"Write"  {set Access "w"}
	"Insert" {set Access "r+"}
	default  {set Access "r"}
  }
  
  # we try to open the file
  if {[catch {open ${FileName} $Access} FileId] } {
    # we can't open this file
    set Local(ErrorMsg) [msgcat::mc "We can't open this file: %s" ${FileName}] 
    set Local(ErrorDetail) $FileId 
    return 1
  } 

  # we read the file in binary format 
  fconfigure $FileId -encoding binary -translation binary

  set Local(FileId) $FileId
  
  return 0
}


###########################################################
# Check the file size of an au-file
###########################################################
proc ::AuFile::FileSize { FileName } {
  variable Local 
  
  # we reset the file size information
  set Local(FileSize) 0
  
  # we check if the file exists 
  if { [file exists ${FileName}] == 0 } {
	set Local(ErrorMsg) [msgcat::mc "This file doesn't exists: %s" ${FileName}] 
	set Local(ErrorDetail) ""
	# we return with an error
	return 1
  }
  
  # ask for the file size (needed in the au-file header)
  set Local(FileSize) [file size ${FileName}]

  # check if the file size is big enough to hold tags
  if {$Local(FileSize) > 552} {
	# this could be a tagged au-file
	return 0
  } else {
	# too small for tags, we test if there enough space for a standard au-file header
	if {$Local(FileSize) > 28} {
      set Local(ErrorMsg) [msgcat::mc "This au-file has no tags: %s" ${FileName}] 
	  set Local(ErrorDetail) [msgcat::mc "The size of the file is too small"]
	} else {
	  # the file exists but it can't never be an au-file
	  set Local(ErrorMsg) [msgcat::mc "This can't be an au-file: %s" ${FileName}] 
	  set Local(ErrorDetail) [msgcat::mc "The size of the file is too small"]
	}
  }
  # if we reach this line: an error occurred before
  return 1
} 


###########################################################
# Check the standard header of an au-file
###########################################################
proc ::AuFile::CheckStdHeader { Header } {
  variable Local

  # we need the offset information for further test
  set Local(StdHeaderOffset) 0 
  
  set Count [binary scan $Header A4I1I1I1I1I1I1 FileType Offset Size Coding SampleRate Channels Info]

  if { $Count != 7 } {
    set Local(ErrorMsg) [msgcat::mc "This can't be an au-file."]
	set Local(ErrorDetail) [msgcat::mc "Not a full au-file header found."]
	return 1
  }
	
  if { $FileType != ".snd" } {
    set Local(ErrorMsg) [msgcat::mc "This can't be an au-file."]
	set Local(ErrorDetail) [msgcat::mc "Wrong file identification field found."]
	return 1
  }

  if { $Coding != 3 } {
    set Local(ErrorMsg) [msgcat::mc "Wrong coding type."]
	set Local(ErrorDetail) [msgcat::mc "We support 16 bit and linear coding only."]
	return 1
  }
 
  if { $SampleRate != 44100 } {
    set Local(ErrorMsg) [msgcat::mc "Wrong sampling rate."]
	set Local(ErrorDetail) [msgcat::mc "We support only 44.1kHz (CD format)."]
	return 1
  }

  if { $Channels != 2 } {
    set Local(ErrorMsg) [msgcat::mc "Wrong count of audio channels."]
	set Local(ErrorDetail) [msgcat::mc "We support stereo only."]
	return 1
  }
  
  # it seems all is okay, we save the offset value
  set Local(StdHeaderOffset) $Offset
  return 0
}


###########################################################
# Precheck the au-file in front of reading from or writing
# to it
###########################################################
proc ::AuFile::FilePreChecks { FileName {Type "Read"} } {
  variable Local 

  # check if the file could have tags
  if { [::AuFile::FileSize ${FileName}] } {
	# there is nothing to do anymore
	return 1
  }
      
  # we try to open the file
  if { [::AuFile::OpenFile ${FileName}] } {
    # the error message is already set
	return 1
  }

  # The first seven uint32 variables are the official
  # header of an au-file
  set AuHeader [read $Local(FileId) 28] 
         
  # check the correctness of the standard header
  if { [::AuFile::CheckStdHeader $AuHeader] } {
    # the error message is already set
	return 1
  }
  
  # if tags exist data have to start after 552 bytes
  if { $Local(StdHeaderOffset) != 552 } {
    set Local(ErrorMsg) [msgcat::mc "File doesn't contain tags."]
	set Local(ErrorDetail) [msgcat::mc "The area for tags is too small."]
	return 1
  }

  # if we intend to write to FileName after this procedure we close the file 
  # channel here and reopen it later for writing again
  if { $Type == "Write" } {
    close $Local(FileId)
  }  

  # if we come until here, all was okay
  return 0
}


###########################################################
# Read the tags from an au-file
###########################################################
proc ::AuFile::ReadTags { FileName } {
  variable Local 

  # reset the tag list
  set TagsList ""
  
  # first we do some file pre-checks
  if { [::AuFile::FilePreChecks ${FileName}] } {
	# an error occurred during pre-checks, we can't go further
	return 1
  }
  
  # we try to read the tags	
  set Tags [read $Local(FileId) 524]
     
  # after reading we can close the file
  close $Local(FileId)
  
  # we search the TAG1 field first
  binary scan $Tags A4 TagId 
  
  if { [string equal $TagId TAG1] } { 
    # we scan again to fetch all information
	set Count [binary scan $Tags A4A128A128A128A128A4A3c TagId Title Artist Album Comment Year Track Genre] 
  } else {
    set Local(ErrorMsg) [msgcat::mc "File doesn't contain tags."]
	set Local(ErrorDetail) [msgcat::mc "We didn't find the key word: TAG1."]
	return 1
  }
  
  # check if we found all fields
  if { $Count != 8 } {
    set Local(ErrorMsg) [msgcat::mc "We couldn't read a complete tag field."]
	set Local(ErrorDetail) [msgcat::mc "We found only %s fields." $Count]
	return 1
  }

  # check if we found a valid genre
  if { $Genre > $Local(GenreCatLen)} {
	# we "limited" the genre value, in the error case we set it to "Other" (13)
	set Genre 13
  }

  # now we compose the list with the success and all tags, we cut all trailing spaces
  set TagsList [list 0 [string trimright $Title] [string trimright $Artist] [string trimright $Album] \
  [string trimright $Comment] [string trimright $Year] [string trimright $Track] \
  [lindex $::AuFile::GenreCat [expr {$Genre - 1}]]]
  
  # and return the success
  return $TagsList  
}


###########################################################
# Write tags into an au-file
###########################################################
proc ::AuFile::AddTags { FileName TagsList } {
  variable Local 

  # first we do some file pre-checks
  if { [::AuFile::FilePreChecks ${FileName} Write] } {
	# an error occurred during pre-checks, we can't go further
	return 1
  } 

  # now we try to open the file
  if { [::AuFile::OpenFile ${FileName} Insert] } {
    # the error message is already set
	return 1
  }

  # we set the file pointer on the first byte
  seek $Local(FileId) 0 start
  
  # first we write the standard header including tag offset
  puts -nonewline $Local(FileId) [binary format a4I1I1I1I1I1I1 ".snd" 552 [expr {$Local(FileSize) - 552}] 3 44100 2 0] 


  # first we convert the genre text into an integer
  set SearchResult [lsearch $::AuFile::GenreCat [lindex $TagsList 6]]
  
  if { $SearchResult == -1 } {
	# the genre wasn't found, we replace it with "Other"
	set Genre 13
  } else {
	# we found it, lists start with zero, genres with 1, we correct it
	set Genre [expr {$SearchResult + 1}]
  }

  # the structure of the TagsList is: Title Artist Album Comment Year Track Genre (in form of text)
  set Tags [binary format A4A128A128A128A128A4A3c "TAG1" [lindex $TagsList 0] [lindex $TagsList 1] \
  [lindex $TagsList 2] [lindex $TagsList 3] [lindex $TagsList 4] [lindex $TagsList 5] $Genre]
  
  puts -nonewline $Local(FileId) $Tags 

  # now we are finished, we can close the file
  close $Local(FileId)

  # and return the success
  return 0
}


###########################################################
# Rip tracks with cdparanoia and write it into an au-file
########################################################### 
proc ::AuFile::RipTracks {CdRomDrive FileName StartTrack EndTrack} {
  variable Local;
  
  # reset the module global error flag
  set Local(ErrorFlag) 0
  set Local(RipFinished) 0
  set Local(RipCounter) 0 

  # first we try to open the output file
  if { [::AuFile::OpenFile ${FileName} Write] } {
    # The error message was already set in OpenFile
	return 1
  }
 
  # Generate the standard sun audio file header (all fields have uint32 data format):
  # 1. value of ".snd" as file tag
  # 2. the offset to the audio data (7x4 bytes standard header + 524 bytes for tags)
  # 3. file size (which is still unknown at this moment, set it to maximum)
  # 4. format (3 = 16bit linear)
  # 5. sample rate
  # 6. number of channels
  # 7. additional information, we set it to zero
  set AuHeader [binary format A4I1H8I1I1I1I1 ".snd" 552 FFFFFFFF 3 44100 2 0]
  
  # write the standard header into the file
  puts -nonewline $Local(FileId) $AuHeader

  # the tags are still unknown, we do some presets
  set Title "Unknown"
  set Artist "Unknown"
  set Album "Unknown"
  set Year [clock format [clock seconds] -format %Y]
  # if we rip track by track: start track and end track are the same
  # if we rip a live record: we have only one file which contains the start
  # track, disadvantage: you rip track 3 to 5 into one file, the title of
  # these three songs is the title of track 3 (you will lose the title of 4 and
  # 5)
  set Track $StartTrack
  set Genre 13

  # Tag definition based on ID3v1 but we spend 128 characters for title, artist, album and comment.
  # We start with the mark "TAG1", year has four characters, track has three.
  # Genre uses the same integer values as ID3v1
  set Tags [binary format A4A128A128A128A128A4A3c "TAG1" $Title $Artist $Album $::AuFile::Comment $Year $Track $Genre]
 
  # write the tag block into the file
  puts -nonewline $Local(FileId) $Tags 
  
  # construct the command for cdparanoia
  if {$Local(OS) == "OpenBSD_Old"} {
	# old OpenBSD's have a problem with the "--" ???
	set Cmd "cdparanoia -q -R -d $CdRomDrive ${StartTrack}-${EndTrack} -"  
  } else { 
	set Cmd "cdparanoia -q -R -d $CdRomDrive ${StartTrack}-${EndTrack} -- -"  
  }

  # execute the command, pass the output of cdparanoia to FidIn
  if { [catch  {open "|$Cmd 2>@ stdout" r} FidIn] } {
    set Local(ErrorMsg) [msgcat::mc "We can't rip the track"] 
	set Local(ErrorDetail) $FidIn
	return 1
  } 
  
  # configure the cdparanoia stream (we call it "input")
  fconfigure $FidIn -translation binary -encoding binary -blocking 0
  # we fetch available data with the CopyStream function
  fileevent  $FidIn readable [list ::AuFile::CopyStream $FidIn $Local(FileId)]

  # we wait here until the track is ripped
  vwait ::AuFile::Local(RipFinished)

  # close the cdparanoia channel
  close $FidIn

  # close the generated au-file
  flush $Local(FileId)
  close $Local(FileId)

  return 0
}


###########################################################
# File event function to copy the data from cdparanoia
# into the au-file
###########################################################
proc ::AuFile::CopyStream {FidIn FidOut} {
   
  # check for end of file in the input channel
  if {[eof $FidIn]} {
    # erase the file event binding
	fileevent $FidIn readable {}
    # set the finished flag
	set ::AuFile::Local(RipFinished) 1
  }

  # Read the input channel
  set Data [read $FidIn]
  # we ignore empty reads
  if {![string length $Data]} return

  # Write data into the au-file
  puts -nonewline $FidOut $Data
}


###########################################################
# Convert an au-file into a headerless raw data file
########################################################### 
proc ::AuFile::AuToRaw {AuFileName RawFileName {BigEndianToLittleEndian 0}} {
  variable Local 
  
  # first we do some file pre-checks
  if { [::AuFile::FilePreChecks ${AuFileName}] } {
	# an error occurred during pre-checks, we can't go further
	return 1
  } 

  # we try to open the raw data file
  if {[catch {open ${RawFileName} w} RawFileId] } {
    # we can't open this file
    set Local(ErrorMsg) [msgcat::mc "We can't open this file: %s" ${RawFileName}] 
    set Local(ErrorDetail) $RawFileId 
    return 1
  } 

  # we write the raw data file in binary format 
  fconfigure $RawFileId -encoding binary -translation binary

  # we go to the first data byte in the au-file
  # (plus one to jump over all headers)
  seek $Local(FileId) [expr {$Local(StdHeaderOffset) + 1}] start

  # the FileSize and Offset are set in FilePreChecks  
  set CntOfRawBytes [expr {$Local(FileSize) - $Local(StdHeaderOffset)}] 
  
  # we copy the raw data in blocks of 65536 bytes
  set BlockSize 65536
  set CopyLoops [expr {int($CntOfRawBytes / $BlockSize)}]
  set BytesLeft [expr {$CntOfRawBytes % $BlockSize}]

  # check if we should convert the big endian format of the au-files
  # into little endian format 
  if { $BigEndianToLittleEndian } {
    # raw file is generate in little endian
	for {set i 0} {$i < $CopyLoops} {incr i} {
	  # Read the block from the au file (in big endian),
	  # convert it to little endian and append it to the raw-file
	  binary scan [read $Local(FileId) $BlockSize] S* Data 
	  puts -nonewline $RawFileId [binary format s* $Data]
    } 
    
	if { $BytesLeft > 0 } {
	  # We read the left bytes out
	  # and put it into the raw-file
	  binary scan [read $Local(FileId) $BytesLeft] S* Data 
      puts -nonewline $RawFileId [binary format s* $Data]
    }
  } else {
    # raw file is big endian also
    for {set i 0} {$i < $CopyLoops} {incr i} {
	  # Read the block from the au file and append it to the raw-file
	  puts -nonewline $RawFileId [read $Local(FileId) $BlockSize]
    }

	if { $BytesLeft > 0 } {
	  # We read the left bytes out
	  # and put it into the raw-file
	  puts -nonewline $RawFileId [read $Local(FileId) $BytesLeft] 
    } 
  }

  # close the au and the raw file
  close $RawFileId 
  close $Local(FileId)

  return 0
}


###########################################################
# Returns the error message and the error details
########################################################### 
proc ::AuFile::GetErrorMsg {} {
  variable Local

 return [list $Local(ErrorMsg) $Local(ErrorDetail)] 
} 

