#*****************************************************************************
# Script rips a CD and save the content into a Sun audio file 
#
# Author: Falk Richter, falk.richter@yandex.com, 2022 
# $Id: cdripper.tcl,v 1.9 2021/07/07 21:32:25 falk Exp falk $
#***************************************************************************** 

namespace eval ::CdRipper {
  variable Local; 
  variable Config;
  # we need only this functions from the outside
  namespace export Main ClearGui

  upvar ::AuTags::CONFIG ::CdRipper::Config
}


####################################################################
# The main proc for the CD-ripper
####################################################################
proc ::CdRipper::Main { Frame OsType } {
  variable Local; 
   
  # we copy the in AuTags generated frame for the ripper
  set Local(cdripperframe) $Frame
  # we copy the type of operation system
  set Local(OS) $OsType
  # the count of the (later generated) track list
  set Local(TrackListItems) 0
  # we activate the checkbuttons
  set Local(SelectAllTracks) 1   
  
  # call the gui
  ::CdRipper::Gui 
}


####################################################################
# After a switch to this program part we make some presettings
####################################################################
proc ::CdRipper::ClearGui {} {
  variable Local
   
  # we destroy all existing widgets in the scroll frame
  if { $Local(TrackListItems) != 0 } {
    foreach Widget [winfo children .$Local(TrackListFrame)] {
      catch {destroy $Widget}
    }
    # reset the count of title fields
    set Local(TrackListItems) 0
  }
  
  # we activate the checkbuttons
  set Local(SelectAllTracks) 1  
  
  # disable the save and the start buttons
  ::FR_general::MakeDisable .$Local(cdripperframe).save .$Local(cdripperframe).start
}


####################################################################
# Fill the combobox for the CD drive with content
####################################################################
proc ::CdRipper::DriveScan {} {
  variable Config; variable Local

  # we use this to set the -state of the scan-button
  set Local(StateScanButton) "disable"

  set DriveList ""
  set DriveCount 0

  # without cdparanoia the search for CD-ROM devices makes no sense
  if {[::AuTags::CheckBackgroundTools Ripper] == ""} {
    # we return here and let the content scan button disabled
	return [msgcat::mc "none"]
  } 

  # cdparanoia exists, we start the search for drives
  ::FR_msgWindow::Add -format Heading -msg [msgcat::mc "We are looking for cd-drives:"]

  switch -- $Local(OS) {
	"OpenBSD_Old" -
	"OpenBSD" {
	  # cdparanoia -s works in OpenBSD but we use dmesg
	  # output here also 
	  if { [catch {exec /sbin/dmesg} CatchOutput] == 0 } {
		# find all lines which start with CD and a number
		set DriveInfo [regexp -line -all -inline {^cd[0-9]+} $CatchOutput]
		# we sort out multiple items
		set DriveInfo [lsort -unique $DriveInfo]
		# set the count of found drives
		set DriveCount [llength $DriveInfo]
		# cdparanoia needs e.g. /dev/cd0c as drive name
		foreach Arg $DriveInfo {
		  lappend DriveList "/dev/${Arg}c"
		}
	  }
	}
	"NetBSD" {
	  # cdparanoia -s doesn't work in NetBSD 1.5.3 yet
	  # we fetch the dmesg output 
	  if { [catch {exec dmesg} CatchOutput] == 0 } {
		# find all lines which start with CD and a number
		set DriveInfo [regexp -line -all -inline {^cd[0-9]+} $CatchOutput]
		# we sort out multiple items
		set DriveInfo [lsort -unique $DriveInfo]
		# set the count of found drives
		set DriveCount [llength $DriveInfo]
		# cdparanoia needs e.g. /dev/cd0a as drive name
		foreach Arg $DriveInfo {
		  lappend DriveList "/dev/${Arg}a"
		}
	  }
	}
	"Linux" {
	  if { [catch {exec lsblk} CatchOutput] == 0 } {
		# find all lines which start with CD and a number
		set DriveInfo [regexp -line -all -inline {^sr[0-9]+} $CatchOutput]
		# we sort out multiple items
		set DriveInfo [lsort -unique $DriveInfo]
		# set the count of found drives
		set DriveCount [llength $DriveInfo]
		# cdparanoia needs e.g. /dev/sr0 as drive name
		foreach Arg $DriveInfo {
		  lappend DriveList "/dev/${Arg}"
		}
	  }
	}
  }

  if {$DriveCount == 0} {
	::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "Sorry, no CD-drives found."]
	set DriveList [msgcat::mc "none"]
  } else {
	::FR_msgWindow::Add -format Body -msg [msgcat::mc "We found %s cdrom drives" $DriveCount] 

	set Local(StateScanButton) "normal"
  }

  return $DriveList
}


####################################################################
# Scan the CD and fill the title count into the combo boxes
####################################################################
proc ::CdRipper::ScanCd {} {
  variable Local; variable Config;
 
  # we start without exceptions...
  set Local(NumOfTracks) 0 
   
  if { [::FR_messageBox::Create -title [msgcat::mc "Content of CD-ROM"] -type Info \
    -message [msgcat::mc "Please put the audio-CD in\nthe CD-ROM drive and press Ok."] \
    -button [list [msgcat::mc "Okay"] [msgcat::mc "Cancel"]] ] == 0 } {
	# the cancel button was pressed
	return
  }
  
  # flag is needed for the output in the message window
  set Failure 0

  ::FR_msgWindow::Add -format Heading -msg [msgcat::mc "We read the content of the CD-ROM:"]
          
  # start cdparanoia and try to read the audio CD
  set RawToc [FR_exec::Start "cdparanoia -Q -d $Config(Device)"] 
  
  # check if the read out was successful 
  if { [lindex $RawToc 0] == 0 } {
	# parse the TOC 
    set ReturnList [::cddb::ProcessToc [lindex $RawToc 1]]
  } else {
    # an error occurred during read out
	::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "The read out failed: %s" [lindex $RawToc 1]]
	# set the failure flag
	set Failure 1
  } 

  # we do this part only if no error occurred until now
  if { $Failure == 0 } {
	# check the success of the raw TOC parsing
	if { [lindex $ReturnList 0] == 0 } {
	  # we handed over the results
	  set Local(TrackList) [lindex $ReturnList 1]
	  set Local(DiscIdString) [lindex $ReturnList 2]
	  # the number of tracks is one part of the disc id string
	  set Local(NumOfTracks) [lindex $Local(DiscIdString) 1]
	} else {
	  # an error occurred, fetch the error form the cddb module 
	  set ErrorList [::cddb::GetErrorMsg]
	  # show it in the message window
	  ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "Processing of the table of content failed: %s" \
		[lindex $ErrorList 0]]
	  ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "Details: %s" \
		[lindex $ErrorList 1]]
	  # set the failure flag
	  set Failure 1
	}
  }

  # we only put out something if no error occurred before
  if { $Failure == 0 } { 
    # check if we found some tracks
	if {$Local(NumOfTracks) > 0} {
      ::FR_msgWindow::Add -format Body -msg [msgcat::mc "We found: %s audio tracks" $Local(NumOfTracks)]
      # fill all found tracks into the scroll frame
      ::CdRipper::AddTracks
	  # activate the select all button
	  .$Local(cdripperframe).selectall configure -state normal
	  # we activate the directory save button
	  .$Local(cdripperframe).save configure -state normal 
    } else {
      ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "We didn't find any audio tracks on this cd."]
    }
  }
}


####################################################################
# add a number for each track into the scroll frame
####################################################################
proc ::CdRipper::AddTracks {} {
  variable Local
 
  # we destroy all existing widgets in the scroll frame
  if { $Local(TrackListItems) != 0 } {
	foreach Widget [winfo children .$Local(TrackListFrame)] {
	  catch {destroy $Widget}
    }
	# reset the count of title fields
	set Local(TrackListItems) 0
  }

  set Index 0

  foreach Track $Local(TrackList) {

	# per default we select all tracks
	set Local(SelectTrack$Index) 1
	checkbutton .$Local(TrackListFrame).select$Index -variable ::CdRipper::Local(SelectTrack$Index)
	label .$Local(TrackListFrame).track$Index -text "$Track"	

	grid .$Local(TrackListFrame).select$Index -row $Index -column 0 -sticky w
	grid .$Local(TrackListFrame).track$Index -row $Index -column 1 -sticky w	

	# we prepare the next index
	incr Index 
  }

  grid columnconfigure .$Local(TrackListFrame) 1 -weight 1

  # we need the count of widgets to destroy all at the next start of AddTracks,
  # we use Index here because it was already increased in the last loop 
  set Local(TrackListItems) $Index 
  
  ::FR_general::ScrollWindowRefresh $Local(TrackChoiceFrame) 
  
  # we configure the scroll increment with help of the first label
  ::FR_general::ScrollWindowIncrement $Local(TrackChoiceFrame) $Local(TrackListFrame).track0 

  # we select all songs as default
  #::CdRipper::SelectAllTracks
} 


####################################################################
# select or deselect all tracks in the scroll window
####################################################################
proc ::CdRipper::SelectAllTracks {} {
  variable Local; 

  # we toggle the flag first
  if {$Local(SelectAllTracks)} {
	set Local(SelectAllTracks) 0
	# and adapt the text of the button
	.$Local(cdripperframe).selectall configure -text [msgcat::mc "Select all tracks"]
  } else {
	set Local(SelectAllTracks) 1
	# and adapt the text of the button
	.$Local(cdripperframe).selectall configure -text [msgcat::mc "Deselect all tracks"]
  } 
  
  for {set Idx 0} {$Idx < $Local(TrackListItems)} {incr Idx} {
   # select or deselect all tracks
   set Local(SelectTrack$Idx) $Local(SelectAllTracks)
  } 
} 


####################################################################
# fetch the selecting tracks in the scroll frame
####################################################################
proc ::CdRipper::TrackSelection {} {
  variable Local

  set Local(SelectedTracks) ""

  # the start index
  set Idx 0

  while {$Idx < $Local(TrackListItems)} {
	# check the state of the checkbutton
	if {$Local(SelectTrack$Idx)} {
	  # the "TrackList" was used to build up the scrolling frame
	  # so we use it here directly again
	  lappend Local(SelectedTracks) [lindex $Local(TrackList) $Idx]                	  	
    }
	# we prepare the next index
	incr Idx
  } 
    
  if { $Local(SelectedTracks) == "" } {
	# no track selected, we don't go further and return here
	::FR_msgWindow::Add -format Error \
    -msg [msgcat::mc "There aren't any tracks selected. I can't do anything for you... :-("]
	
	return 1
  }

  # for ripping multiple tracks into one file (e.g. for live records)
  if {$Local(LiveRecording) == 1} {
	# we need a continuously increasing of the track index
	set TracksCnt [expr {([lindex $Local(SelectedTracks) end] - [lindex $Local(SelectedTracks) 0]) + 1}]
	# e.g. you selected 7 8 9 10 -> 10 - 7 + 1 = 4 -> okay
	# but 7 8 10 -> also 4 but only 3 tracks -> not okay
	if {$TracksCnt != [llength $Local(SelectedTracks)]} {
      ::FR_msgWindow::Add -format Error \
      -msg [msgcat::mc "You selected: \"All tracks into one file\". \
A continuously increasing index of tracks is required for this mode (e.g. track 7, 8, 9 but not 7, 9, 13)."]
	  # we erase the list again
      set Local(SelectedTracks) ""
	  return 1
	}
  } 

  ::FR_msgWindow::Add -format Info -msg [msgcat::mc "You selected track(s): %s" $Local(SelectedTracks)]
  
  return 0 
} 


####################################################################
# The "save where" button
####################################################################
proc ::CdRipper::SaveOutputFiles {} {
  variable Local; variable Config
  
  set OutputDir [::FR_dirBox::Create -initialdir [lindex $Config(OutputDir) 0] -permission "writable" \
  -title [msgcat::mc "Choose the directory for your output files:"]]
  
  # check if the return value is not empty 
  if {[lindex $OutputDir 0] != ""} { 
    # copy the new output directory (in list format)
    set Config(OutputDir) $OutputDir
   
    ::FR_msgWindow::Add -format Info \
    -msg [msgcat::mc "The outputfile directory is set to %s" [lindex $Config(OutputDir) 0]]
    # we ready now for ripping
    .$Local(cdripperframe).start configure -state normal
  } 
}


####################################################################
# we create and place all gui elements
####################################################################
proc ::CdRipper::Gui {} {
  variable Config; variable Local
 
  # the widgets
  label .$Local(cdripperframe).device_comment -pady 10 -text [msgcat::mc "Ripping-device:"]
  ::FR_comboBox::cb .$Local(cdripperframe).device -variable ::CdRipper::Config(Device) -entry_width auto \
  -listbox_content [::CdRipper::DriveScan] -default_value $::CdRipper::Config(Device)
  
  label .$Local(cdripperframe).scan_comment -pady 10 -text [msgcat::mc "Scan the content of the CD (TOC):"]
  button .$Local(cdripperframe).scan -text [msgcat::mc "Start scan"] \
  -state $::CdRipper::Local(StateScanButton) -command {::CdRipper::ScanCd}
 
  label .$Local(cdripperframe).trackchoice_comment -pady 10 -text [msgcat::mc "Select the titles which you want to rip:"]
  
  # the widgets to choice the tracks
  set Local(TrackChoiceFrame) "$Local(cdripperframe).trackchoice"
  frame .$Local(TrackChoiceFrame) -relief flat

  button .$Local(cdripperframe).selectall -text [msgcat::mc "Deselect all tracks"] \
  -command ::CdRipper::SelectAllTracks -state disabled
 
  checkbutton .$Local(cdripperframe).live -pady 10 -variable ::CdRipper::Local(LiveRecording) \
  -text [msgcat::mc "All selected tracks in one file (e.g. for a live record)?"]
 
  label .$Local(cdripperframe).save_comment -pady 10 -text [msgcat::mc "Save the output files to:"]
  button .$Local(cdripperframe).save -text [msgcat::mc "Save"] -state disable \
  -command {::CdRipper::SaveOutputFiles}
 
  button .$Local(cdripperframe).start -text [msgcat::mc "Start ripping"] -foreground blue -state disable \
  -command {::CdRipper::CdripperStartButton}
 
  # the placement
  # the cdrom device
  grid .$Local(cdripperframe).device_comment -row 0 -column 0 -columnspan 3 -sticky w
  grid .$Local(cdripperframe).device -row 0 -column 4 -sticky e
  # a line at row 1 over 5 columns
  ::FR_general::Line $Local(cdripperframe) 1 5
  # the scan button
  grid .$Local(cdripperframe).scan_comment -row 2 -column 0 -columnspan 3 -sticky w
  grid .$Local(cdripperframe).scan -row 2 -column 4 -sticky ew
  # a line at row 3 over 5 columns 
  ::FR_general::Line $Local(cdripperframe) 3 5
 
  # the headline over the title select frame 
  grid .$Local(cdripperframe).trackchoice_comment -row 4 -column 0 -columnspan 3 -sticky w

  grid .$Local(TrackChoiceFrame) -row 5 -column 0 -sticky w
 
  grid .$Local(cdripperframe).selectall -row 5 -column 4 -sticky ew
 
  # back to the ripper frame...the "live" checkbutton
  grid .$Local(cdripperframe).live -row 6 -column 0 -sticky w
  
  # again a line at row 7 over 5 columns
  ::FR_general::Line $Local(cdripperframe) 7 5
  # the save directory 
  grid .$Local(cdripperframe).save_comment -row 8 -column 0 -columnspan 3 -sticky w 
  grid .$Local(cdripperframe).save -row 8 -column 4 -sticky ew
 
  ::FR_general::Line $Local(cdripperframe) 9 5
  # the pad frame to fill the empty space
  ::FR_general::PadFrame $Local(cdripperframe) 10 5
  grid .$Local(cdripperframe).start -row 11 -column 0 -columnspan 5 -sticky ew
 
  grid columnconfigure .$Local(cdripperframe) 3 -weight 1

  # the scroll window with a height of 2cm and a width of 1.5cm
  set Local(TrackListFrame) [::FR_general::ScrollWindow $Local(TrackChoiceFrame) 2c 1.5c]

  # the balloon help
  ::FR_balloonHelp::AddHelp ".$Local(cdripperframe).device" [msgcat::mc "Device name of your CD-ROM drive"]
  ::FR_balloonHelp::AddHelp ".$Local(cdripperframe).scan" [msgcat::mc "Scan the content of the audio-CD first"]
  ::FR_balloonHelp::AddHelp ".$Local(cdripperframe).save" \
  [msgcat::mc "This is the directory in which you want to rip the CD-tracks"]
  ::FR_balloonHelp::AddHelp ".$Local(cdripperframe).trackchoice" [msgcat::mc "Select the desired tracks"]
  ::FR_balloonHelp::AddHelp ".$Local(cdripperframe).live" \
  [msgcat::mc "We rip all chosen tracks in one file (without interruptions)"]
 
  ::AuTags::BalloonHelpStatus
}


####################################################################
# The start button
####################################################################
proc ::CdRipper::CdripperStartButton {} {
  variable Local; variable Config
 
  # we prepare it for the next CD
  ::FR_general::MakeDisable .$Local(cdripperframe).selectall .$Local(cdripperframe).save .$Local(cdripperframe).start
  
  # start the check for the selected tracks
  if {[::CdRipper::TrackSelection]} {
	return 1
  } 
 
  # we go to the selected output directory
  cd [lindex $Config(OutputDir) 0]
   
  # first we create unique file names for the ripped files
  set ReturnValue [::CdRipper::CreateFileNames]
   
  # check if there was no "abort" during creating the file names 
  if {$ReturnValue == 0} {
    # we try to start the ripping
    set ReturnValue [::CdRipper::Rip]
  }
   
  # check the rip results
  if {$ReturnValue == 0} {
	# Create a file with the information for the gnudb database
    set ReturnValue [::CdRipper::WriteGdbFile]
  }

  # check the results
  if {$ReturnValue == 0} {	
    ::FR_messageBox::Create -title [msgcat::mc "End of job"] -type Info \
    -message [msgcat::mc "We have ripped all tracks successfully"]
  } else {
    ::FR_messageBox::Create -type Error -message \
    [msgcat::mc "We couldn't finish this job successfully.\n\nPlease check the message window for further information."]
  }   
 
  # we prepare the track list for the next scan...
  ::CdRipper::ClearGui 
}


####################################################################
# We check for existing files and create a file list
####################################################################
proc ::CdRipper::CreateFileNames {} {
  variable Config; variable Local

  set ExistingFiles [::AuTags::CheckDirContent [lindex $Config(OutputDir) 0] $Config(OutputFormat)]
 
  if {$ExistingFiles == ""} {
    # the directory don't contain any files in Config(OutputFormat)
	set Index 0
    set ExistingFiles 0
  } else {
    # in the first step we assume: the existing files are named with index 1, 2, 3...
	set Index [llength $ExistingFiles]
    set ExistingFiles 1
  }
  
  set Local(OutputFiles) ""
  set Local(Tracks) ""
  
  # we do it for each selected track in the listbox
  foreach Track $Local(SelectedTracks) {
    # generate a new index
	incr Index
    # create the new file name
    set OutputFileName [::CdRipper::EditIndex $Index]  
	# test it
    while {[file exists $OutputFileName]} {
      # ...or correct it...
	  incr Index
      #... until a unique name is found
      set OutputFileName [::CdRipper::EditIndex $Index]
    } 
  
    lappend Local(Tracks) $Track
    lappend Local(OutputFiles) $OutputFileName
  } 
  
  if {$ExistingFiles} {
    set Answer [::FR_messageBox::Create  -title [msgcat::mc "Other files found"] -type Info \
    -button [list [msgcat::mc "Accept"] [msgcat::mc "Abort"]] \
    -message [msgcat::mc "There are already %s-files in the output directory.\n\nWe have looked for \
unused file names and suggest the following names:\n\n%s" $Config(OutputFormat) $Local(OutputFiles)]]
   
    if {$Answer == 0} {
      ::FR_msgWindow::Add -format Error -msg [msgcat::mc "Abort was your decision, we stoped the job."] 
      return 1
    }
  }
 
  return 0
}


####################################################################
# if wished: add a zero in front of the single digit number
####################################################################
proc ::CdRipper::EditIndex {Index} {
  variable Config

  # check if the leading zero is wished
  if { $Config(LeadingZeroAtTrackNumber) } {
	# convert the incex (track number) to integer
	set Index [expr {int($Index)}]
	# check if the track number is a single digit number
	if { $Index < 10 } {
	  return "Song0${Index}.au"
    }
  }
  
  # in all other cases (index is not a single digit number
  # or the leading zero is not wished): we return the index "as is"
  return "Song${Index}.au" 
} 


####################################################################
# The ripping procedures
####################################################################
proc ::CdRipper::Rip {} {
  variable Config; variable Local
 
  set ReturnValue 1

  #  now we can really start to generate au file(s)
  if {$Local(LiveRecording) == 1} { 
	# we rip multiple tracks into one file in this mode
	set ReturnValue [::CdRipper::RipLiveTrack [lindex $Local(Tracks) 0] \
	[lindex $Local(Tracks) end] [lindex $Local(OutputFiles) 0]]
  } else {
    # standard: rip track for track into separate files
	foreach Track $Local(Tracks) OutputFileName $Local(OutputFiles) {
      set ReturnValue [::CdRipper::RipSingleTrack $Track $OutputFileName]
      # we stop immediately after an error 
      if {$ReturnValue} {
        break;
      }
    }
  } 
  return $ReturnValue
}


####################################################################
# sub procedure for tracks ripping
####################################################################
proc ::CdRipper::RipSingleTrack {Track FileName} {
  variable Config;
 
  ::FR_msgWindow::Add -format Heading -msg [msgcat::mc "We rip the %s. track!" $Track]
 
  # check if an error occurred during ripping
  if { [::AuFile::RipTracks $Config(Device) $FileName $Track $Track] } {
    # we fetch the error message
    set RipError [::AuFile::GetErrorMsg]
 
    ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "An error occurred:\n%s" [lindex $RipError 0]]
    ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "Details:\n%s" [lindex $RipError 1]]
    return 1 
  } else { 
    # au-file successfully created
    ::FR_msgWindow::Add -format Body -msg [msgcat::mc "Track successfully ripped!"]
    return 0
  } 
}


####################################################################
# sub procedure for live record ripping
####################################################################
proc ::CdRipper::RipLiveTrack {StartTrack LastTrack FileName} {
 variable Config;
 
  ::FR_msgWindow::Add -format Heading \
  -msg [msgcat::mc "We rip from track %s to %s without gaps!" $StartTrack $LastTrack]

  # check if an error occurred during ripping
  if { [::AuFile::RipTracks $Config(Device) $FileName $StartTrack $LastTrack] } {
    # we fetch the error message
    set RipError [::AuFile::GetErrorMsg]
 
    ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "An error occurred:\n%s" [lindex $RipError 0]]
    ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "Details:\n%s" [lindex $RipError 1]]
    return 1 
  } else {
    # all files successfully created
    ::FR_msgWindow::Add -format Body -msg [msgcat::mc "Tracks successfully ripped!"]
    return 0
  } 
} 


####################################################################
# Write the ID for the gnudb access into a text file
####################################################################
proc ::CdRipper::WriteGdbFile {} {
  variable Local; variable Config;
  
  if {[catch {open $Config(DiscIdFile) w} FileId] } {
    # we can't open this file
    set Local(ErrorMsg) [msgcat::mc "We can't write the file with the Disc id."] 
    set Local(ErrorDetail) $FileId 
    return 1
  }

  # write the ID string needed by the gnudb into the text file
  puts $FileId $Local(DiscIdString)

  # we close the file again
  close $FileId
  
  return 0
}

