#*****************************************************************************
# Script converts (extended) Sun audio files into mp3 using lame
#
# Author: Falk Richter, falk.richter@yandex.com, 2021 
# $Id: converter.tcl,v 1.4 2021/12/08 20:23:11 falk Exp falk $
#***************************************************************************** 

namespace eval ::AuConverter {
  variable Local; 
  variable Config;
  # we need only this functions from the outside
  namespace export Main ClearGui

  upvar ::AuTags::CONFIG ::AuConverter::Config
}


####################################################################
# The main proc for the converter
#################################################################### 
proc ::AuConverter::Main { Frame } {
  variable Local
  
  # we copy the in AuTags generated frame for the converter
  set Local(converterframe) $Frame 
  
  # the count of the (later generated) file list
  set Local(FileListItems) 0
  
  # we deactivate the checkbutton
  set Local(SelectAllSongs) 0 
  
  # the list with all files of the chosen directory
  set Local(AllFiles) ""
  
  # Create the gui
  ::AuConverter::Gui

  # lame is needed
  if {[::AuTags::CheckBackgroundTools Mp3Converter] != ""} {
    # we activate the au-file directory button only if lame
	# is available
	.$Local(converterframe).filedir configure -state normal 
  } 
}


####################################################################
# after a switch to this program part we make some presettings
####################################################################
proc ::AuConverter::ClearGui {} {
  variable Local 
  
  # we destroy all existing widgets in the scroll frame
  if { $Local(FileListItems) != 0 } {
    # there exists title fields
	foreach Widget [winfo children .$Local(TitleFrame)] {
      catch {destroy $Widget}
    }
    # reset the count of title fields 
    set Local(FileListItems) 0
  }

  # we deactivate the checkbutton
  set Local(SelectAllSongs) 0 
  
  .$Local(converterframe).mp3dir configure -state disabled
  .$Local(converterframe).filelist_selectall configure -state disabled
  .$Local(converterframe).convert configure -state disabled
}


####################################################################
# we create and place all gui elements
####################################################################
proc ::AuConverter::Gui {} {
  variable Local; variable Config
  
  # the widgets 
  label .$Local(converterframe).filedir_comment  -pady 10 \
  -text [msgcat::mc "Choose the directory with your au-files:"]
  button .$Local(converterframe).filedir -text [msgcat::mc "Open"] -state disabled \
  -command {::AuConverter::FileDir} 
  
  label .$Local(converterframe).mp3dir_comment  -pady 10 \
  -text [msgcat::mc "Choose the directory for the mp3-files:"]
  button .$Local(converterframe).mp3dir -text [msgcat::mc "Open"] -state disabled \
  -command {::AuConverter::Mp3Dir} 

  button .$Local(converterframe).filelist_selectall -command ::AuConverter::SelectAllSongs -pady 5 \
  -text [msgcat::mc "Select all songs"] -state disabled 
  
  set Local(FileListFrame) "$Local(converterframe).filelist"
  frame .$Local(FileListFrame) -relief flat

  button .$Local(converterframe).convert -pady 5 -text [msgcat::mc "Convert"] -foreground blue \
  -activeforeground blue -state disabled -command {::AuConverter::ConvertButton}
 
  button .$Local(converterframe).stop -pady 5 -text [msgcat::mc "Stop"] -command {::AuConverter::StopButton} 
  
  # the placement 
  grid .$Local(converterframe).filedir_comment -row 0 -column 0 -sticky w 
  grid .$Local(converterframe).filedir -row 0 -column 2 -sticky ew
  
  ::FR_general::Line $Local(converterframe) 1 3 

  grid .$Local(converterframe).mp3dir_comment -row 2 -column 0 -sticky w
  grid .$Local(converterframe).mp3dir -row 2 -column 2 -sticky ew
  
  ::FR_general::Line $Local(converterframe) 3 3 
  
  grid .$Local(converterframe).filelist_selectall -row 4 -column 0 -sticky w
  grid .$Local(FileListFrame) -row 5 -column 0 -columnspan 3 -sticky nsew -pady 5 
  grid .$Local(converterframe).stop -row 6 -column 0 -sticky ew 
  grid .$Local(converterframe).convert -row 6 -column 1 -columnspan 2 -sticky ew 
  
  grid columnconfigure .$Local(converterframe) 1 -weight 1
  grid rowconfigure .$Local(converterframe) 5 -weight 1 
 
  # the scroll window
  set Local(TitleFrame) [::FR_general::ScrollWindow $Local(FileListFrame)] 
  
  ::FR_balloonHelp::AddHelp ".$Local(converterframe).filedir" \
  [msgcat::mc "This is the directory with your Sun audio files"]
  
  :::FR_balloonHelp::AddHelp ".$Local(converterframe).mp3dir" \
  [msgcat::mc "This is the directory for your mp3 files"]
  
  ::AuTags::BalloonHelpStatus
}


####################################################################
# load the directory for the mp3-files
####################################################################
proc ::AuConverter::Mp3Dir {} {
  variable Config; variable Local

  set Mp3Dir [::FR_dirBox::Create -initialdir [lindex $Config(Mp3Dir) 0] -parent .$Local(converterframe) \
  -title [msgcat::mc "Choose the directory for your mp3-files:"] \
  -filetype [list [msgcat::mc "Mp3 files"] "*.mp3"]]
  
  # check if the answer is empty
  if {[lindex $Mp3Dir 0] != ""} {
	# copy the new directory (in list format)
    set Config(Mp3Dir) $Mp3Dir 

    ::FR_msgWindow::Add -format Info \
	-msg [msgcat::mc "The mp3-directory was set to: %s" [lindex $Config(Mp3Dir) 0]] 

    # now we can activate the convert button
   .$Local(converterframe).convert configure -state normal 
  } 
} 


####################################################################
# load the directory with the au-files
####################################################################
proc ::AuConverter::FileDir {} {
  variable Config; variable Local

  # output directory because we use the output of the tagger or cdripper sub-programs
  set OutputDir [::FR_dirBox::Create -initialdir [lindex $Config(OutputDir) 0] -parent .$Local(converterframe) \
  -title [msgcat::mc "Choose the directory with your au-files:"] -filetype [list [msgcat::mc "Sun audio files"] "*.au"]]
  
  # check if the answer is empty
  if {[lindex $OutputDir 0] != ""} {
	# copy the new directory (in list format)
    set Config(OutputDir) $OutputDir
    
    cd [lindex $Config(OutputDir) 0]

    ::FR_msgWindow::Add -format Info \
	-msg [msgcat::mc "The directory was set to: %s" [lindex $Config(OutputDir) 0]]
    
    ::AuConverter::TitleFieldsCreate
  }
}


####################################################################
# select or deselect all songs in the title list
####################################################################
proc ::AuConverter::SelectAllSongs {} {
  variable Local; 

  # we toggle the flag first
  if {$Local(SelectAllSongs)} {
	set Local(SelectAllSongs) 0
	# and adapt the text of the button
	.$Local(converterframe).filelist_selectall configure -text [msgcat::mc "Select all songs"]
  } else {
	set Local(SelectAllSongs) 1
	# and adapt the text of the button
	.$Local(converterframe).filelist_selectall configure -text [msgcat::mc "Deselect all songs"]
  } 
  
  # the song index starts with zero
  for {set Idx 0} {$Idx < $Local(FileListItems)} {incr Idx} {
   # select or deselect all tracks
   set Local(SelectSong$Idx) $Local(SelectAllSongs)
  } 
}


####################################################################
# the convert button
####################################################################
proc ::AuConverter::ConvertButton {} {
  variable Local
  
  # we start with minus one here because the file list starts with index 0
  # so we can find in FindNextSong the first song
  set Local(CurrentSong) -1
  
  # in case of an error we stop the loop over all selected songs
  set RunLoop 1
  
  # we try to convert all selected songs
  while { $RunLoop } {
    set NextSong [::AuConverter::FindNextSong $Local(CurrentSong)]

	# we separate the index from the return list
    set Idx [lindex $NextSong 0]
	
	if {$Idx == -1} {
	  # no further songs found we stop the while loop
	  break
	} 
	  
	# we found a next song
	set Local(CurrentSong) $Idx 
	
	# start converting
	if {[::AuConverter::Convert [lindex $NextSong 1]]} {
	  # we can not convert the file, we stop the while loop also
	  set RunLoop 0
    }

	# leave the function with a deselected checkbutton
    set Local(SelectSong$Local(CurrentSong)) 0	
	
    # check if the stop button was pressed
	if {$Local(Done) == 2} {
	  # we stop the convert process for all further songs
	  set RunLoop 0
	}
  } 
}


####################################################################
# the stop button
####################################################################
proc ::AuConverter::StopButton {} {
  variable Local 
  
 # we set the finished flag in the convert procedure
  set Local(Done) 2 

  # Put out a message
  ::FR_msgWindow::Add -format Body -msg [msgcat::mc "We stoped the converting process."]
}


####################################################################
# Find the next selected song
####################################################################
proc ::AuConverter::FindNextSong {StartIdx} {
  variable Local
  
  set Found 0 
  
  # the start index is the currently converted song
  # we try the next song
  set Idx [expr {$StartIdx + 1}]

  # the song index starts with zero
  while {$Idx < $Local(FileListItems)} {
    if {$Local(SelectSong$Idx)} {
	  set Found 1
	  break
	}
    incr Idx
  }
  
  if {$Found} {
	# we return the found index (for the next search) and the file name of the song
	return [list $Idx [lindex $Local(AllFiles) $Idx]]
  } else {
	# we didn't find anything, we return an invalid index
	return -1
  }
} 


####################################################################
# convert the handed over song
####################################################################
proc ::AuConverter::Convert {FileName} {
  variable Local; variable Config

  # we use this variable to detect the end of convert process
  set Local(Done) 0

  ::FR_msgWindow::Add -format Heading \
  -msg [msgcat::mc "We convert the %s. track!" [expr {$Local(CurrentSong) + 1}]] 
  
  # restore the full au-file name
  set AuFileName [list [file join [lindex $Config(OutputDir) 0] $FileName]]
  
  # extract the rootname of the file
  set FileRootName [file rootname $FileName]

  # generate the file name for the temporary raw file
  set RawFileName [list [file join [lindex $Config(OutputDir) 0] ${FileRootName}.raw]]

  # generate the full name for the mp3 file
  set Mp3FileName [list [file join [lindex $Config(Mp3Dir) 0] ${FileRootName}.mp3]] 

  ::FR_msgWindow::Add -format Body -msg [msgcat::mc "We prepare the data for the mp3 converter!"] 

  # we convert the au-file into a headerless raw data file
  if { [::AuFile::AuToRaw $AuFileName $RawFileName] } {
	# we fetch the error message
    set Error [::AuFile::GetErrorMsg]
 
    ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "An error occurred:\n%s" [lindex $Error 0]]
    ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "Details:\n%s" [lindex $Error 1]]
    return 1 
  } 
  
  # select the tags of the current song
  set TagList [lindex $Local(AllTags) $Local(CurrentSong)]
  
  # prepare the lame command
  set Cmd "lame -r -s 44.1 -m j -x "
  append Cmd "--tt \"[lindex $TagList 0]\" --ta \"[lindex $TagList 1]\" --tl \"[lindex $TagList 2]\" "
  append Cmd "--tc \"[lindex $TagList 3]\" --ty \"[lindex $TagList 4]\" --tn \"[lindex $TagList 5]\" "
  append Cmd "--tg \"[lindex $TagList 6]\" [lindex $RawFileName 0] [lindex $Mp3FileName 0]" 

  # we direct the error messages also to the standard output
  # (the channel becomes so readable in case of success or error)
  if { [catch { open "|$Cmd 2>@ stdout" r} Local(FidLame)] } {
	::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "We could not start converting song: %s" $FileName] 
	return 1
  } 
  
  # save the program id for the stop function
  set Pid [pid $Local(FidLame)]
  
  # set the channel into the non-blocking mode (for the stop button)
  fconfigure $Local(FidLame) -blocking 0
  
  # place a message with the current state
  ::FR_msgWindow::Add -format Body -msg [msgcat::mc "We convert the au-file into mp3:"] 
  
  # fetch all messages from lame and put it into the message window
  fileevent $Local(FidLame) readable {::AuConverter::ConverterOutput} 
  	
  # we wait until the variable is set
  tkwait variable ::AuConverter::Local(Done)
  
  # close the channel
  fileevent $Local(FidLame) readable {}
  catch {close $Local(FidLame)} 
  
  # in case of the stop function we have to kill the background task also
  if { $Local(Done) == 2 } {
    catch {exec kill $Pid} 
  }

  # remove the temporary raw file
  catch {file delete [lindex $RawFileName 0]}
  
  return 0 
}


####################################################################
# fetch all messages from the converter tool
####################################################################
proc ::AuConverter::ConverterOutput {} { 
  variable Local 
  
  # Read a line if available
  if {[gets $Local(FidLame) Line] > 0} {
    ::FR_msgWindow::Add -format Body -msg "$Line"
  } else {
    # the tool has finished
    if {[eof $Local(FidLame)]} {
	  set Local(Done) 1
    }
  } 
}



####################################################################
# generate one line per song and fill the entry-widgets with the 
# title/artist information
####################################################################
proc ::AuConverter::TitleFieldsCreate {} {
  variable Config; variable Local
  
  # we destroy all existing widgets in the scroll frame
  if { $Local(FileListItems) != 0 } {
    foreach Widget [winfo children .$Local(TitleFrame)] {
      catch {destroy $Widget}
    }
    # reset the count of title fields
    set Local(FileListItems) 0
  }

  # reset the list with all tags
  set Local(AllTags) ""
  
  set Local(AllFiles) [::AuTags::CheckDirContent [lindex $Config(OutputDir) 0] "SunAudio"]

  # check if we found something
  if {$Local(AllFiles) == ""} {
	# deactivate the convert button
    .$Local(converterframe).convert configure -state disabled
    .$Local(converterframe).filelist_selectall configure -state disabled
    
	::FR_msgWindow::Add -format Error \
	-msg [msgcat::mc "I didn't find any au-files. Have you chosen the right directory?"]
    # we can stop here
	return
  } 
   
  # the labels in the head of the table
  label .$Local(TitleFrame).file_comment -text [msgcat::mc "Filename"]
  label .$Local(TitleFrame).artist_comment -text [msgcat::mc "Artist"]
  label .$Local(TitleFrame).title_comment -text [msgcat::mc "Title"]

  # we place the header
  grid .$Local(TitleFrame).file_comment -row 0 -column 1 -sticky w
  grid .$Local(TitleFrame).artist_comment -row 0 -column 2 -sticky w 
  grid .$Local(TitleFrame).title_comment -row 0 -column 3 -sticky w
	
  # the start index
  set Index 0
    
  # now we process each file
  foreach File $Local(AllFiles) { 
    # we limit the count of characters in the file name to 20
    if { [string length $File] > 20 } {
	  # split file name into name and extension
      set RootName [file rootname $File]
      set FileExtension [file extension $File]
      # range from 0...18 are 19 characters + "~" = 20
      # we reduce the file name length by the extension length
	  set NameLimit [expr {18 - [string length $FileExtension]}]
      # we add all together again
	  set FileName [string range $RootName 0 $NameLimit]
      append FileName "~" $FileExtension
    } else {
	  # the standard case: we can show the full file name
	  set FileName $File
    }
   
    # the file depending widgets
	set Local(SelectSong$Index) 0
	checkbutton .$Local(TitleFrame).select$Index -variable ::AuConverter::Local(SelectSong$Index)
	label .$Local(TitleFrame).file$Index -text "$FileName"
    entry .$Local(TitleFrame).artist$Index -state normal -width 28 
	entry .$Local(TitleFrame).title$Index -state normal -width 28

	# the first index is zero but the first row is the header, so the row is index plus one
	set Row [expr {$Index + 1}]
	grid .$Local(TitleFrame).select$Index -row $Row -column 0 -sticky w
	grid .$Local(TitleFrame).file$Index -row $Row -column 1 -sticky w
	grid .$Local(TitleFrame).artist$Index -row $Row -column 2 -sticky ew
	grid .$Local(TitleFrame).title$Index -row $Row -column 3 -sticky ew
	 
    # now we read the tags from the au-files
	set TagList [::AuFile::ReadTags [file join [lindex $Config(OutputDir) 0] $File]]
     
    # we check the success
	if { [lindex $TagList 0] != 0 } {
      # oops an error occurred - we ask for the reason
	  set ErrorList [::AuFile::GetErrorMsg]
    
	  ::FR_msgWindow::Add -format Error -msg [msgcat::mc "An error occurred during reading of file: %s" $File]
      ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "The reason: %s" [lindex $ErrorList 0]] 
      ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "Details: %s" [lindex $ErrorList 1]]

	  # we need the count of widgets to destroy all at the next start of titleartist,
      # Index is zero based so we use Row because it is Index + 1
	  set Local(FileListItems) $Row
    
      return
    }

    # we store all tags (with exception of the Success field) to handed over it later to lame
	lappend Local(AllTags) [lrange $TagList 1 end]
	
	# we generate lists with the information of all au-files
	# the form: Success Title Artist Album Comment Year Track Genre (in form of text) 
    .$Local(TitleFrame).title$Index insert end [lindex $TagList 1]
    .$Local(TitleFrame).artist$Index insert end [lindex $TagList 2] 

    # we prepare the next index
	incr Index 
  }

  grid columnconfigure .$Local(TitleFrame) 2 -weight 1
  grid columnconfigure .$Local(TitleFrame) 3 -weight 1

  # we need the count of widgets to destroy all at the next start of titleartist,
  # we use Index here because it was already increased in the last loop
  set Local(FileListItems) $Index 
  
  ::FR_general::ScrollWindowRefresh $Local(FileListFrame) 
  
  # we configure the scroll increment with help of the first entry-box
  ::FR_general::ScrollWindowIncrement $Local(FileListFrame) $Local(TitleFrame).artist0
  
  .$Local(converterframe).mp3dir configure -state normal
  .$Local(converterframe).filelist_selectall configure -state normal

  # we select all songs as default
  ::AuConverter::SelectAllSongs
}

