#*****************************************************************************
# Script plays (extended) Sun audio files 
#
# Author: Falk Richter, falk.richter@yandex.com, 2021 
# $Id: player.tcl,v 1.7 2021/12/09 20:45:00 falk Exp falk $
#***************************************************************************** 

namespace eval ::AuPlayer {
  variable Local; 
  variable Config;
  # we need only this functions from the outside
  namespace export Main ClearGui

  upvar ::AuTags::CONFIG ::AuPlayer::Config
}


####################################################################
# The main proc for the au-player
#################################################################### 
proc ::AuPlayer::Main { Frame OsType } {
  variable Local

  # we copy the in AuTags generated frame for the player
  set Local(playerframe) $Frame

  # we copy the type of operation system
  set Local(OS) $OsType

  # the count of the (later generated) file list
  set Local(FileListItems) 0

  # we deactivate the checkbuttons
  set Local(SelectAllSongs) 0 

  # the list with all files of the chosen directory
  set Local(AllFiles) ""

  # Create the gui
  ::AuPlayer::Gui

  # define the audio mixer
  switch -- $Local(OS) {
	"OpenBSD" {
	  # later we want to change the parameter output.level,
	  # we list all parameters in the debug mode
	  set Cmd "sndioctl -d"
	  # we check if it exists 
	  if { [catch {exec $Cmd} ExecOutput] == 0 } {
		# look for e.g. "output[0].level"
		if {![regexp {output\[[0-9]+\].level} $ExecOutput]} {
		  # there ist no output[0].level -> we can't control the volume
		  .$Local(playerframe).volume -state disabled
	    }
	  }
	}
	"NetBSD" -
	"OpenBSD_Old" {
	  # both OS use mixerctl to adjust the loudness,
	  # later we want to change the outputs.master level
	  set Cmd "mixerctl outputs.master"
	  # we check if it exists 
	  if { [catch {exec $Cmd} ExecOutput] == 0 } {
		# if there is no "outputs.master" mixerctl would throw
		# an error starting with "mixerctl: ..."
		if {[regexp {^mixerctl:} $ExecOutput]} {
		  # there ist no outputs.master -> we can't control the volume
		  .$Local(playerframe).volume -state disabled
	    }
	  } 
    }
    "Linux" {
	  # later we want to change the 'Master' level as a simple control element
	  set Cmd "amixer scontrols"
	  # we list all simple controls
	  if { [catch {exec $Cmd} ExecOutput] == 0 } {
	    # we are looking for 'Master'
	    if {[regexp {'Master',} $ExecOutput]} {
		  # there is no 'Master' -> we can't control the volume
		  .$Local(playerframe).volume -state disabled
	    }
      } 
    }
  }

  # we check the player tool
  if {[::AuTags::CheckBackgroundTools Player] != ""} {
    # you can select a directory only if the player tool is available
    .$Local(playerframe).filedir configure -state normal
  }
}


####################################################################
# after a switch to this program part we make some presettings
####################################################################
proc ::AuPlayer::ClearGui {} {
  variable Local 
  
  # we destroy all existing widgets in the scroll frame
  if { $Local(FileListItems) != 0 } {
    # there exists title fields
	foreach Widget [winfo children .$Local(TitleFrame)] {
      catch {destroy $Widget}
    }
    # reset the count of title fields 
    set Local(FileListItems) 0
  }

  set Local(SelectAllSongs) 0 
  
  .$Local(playerframe).play configure -state disabled
}


####################################################################
# we create and place all gui elements
####################################################################
proc ::AuPlayer::Gui {} {
  variable Local; variable Config
  
  # the widgets 
  label .$Local(playerframe).filedir_comment  -pady 10 \
  -text [msgcat::mc "Choose the directory with your au-files:"]
  button .$Local(playerframe).filedir -text [msgcat::mc "Open"] -command {::AuPlayer::FileDir} \
  -state disabled
  
  # we put the label for the next frame here
  button .$Local(playerframe).filelist_selectall -command ::AuPlayer::SelectAllSongs -pady 5 \
  -text [msgcat::mc "Select all songs"] -state disabled
  
  label .$Local(playerframe).volume_comment -pady 10 -text [msgcat::mc "Volume control:"]
  scale .$Local(playerframe).volume -from 0 -to 100 -orient horizontal -variable ::AuPlayer::Config(Volume) \
  -command {::AuPlayer::SetVolume}
  
  set Local(FileListFrame) "$Local(playerframe).filelist"
  frame .$Local(FileListFrame) -relief flat 

  button .$Local(playerframe).stop -text [msgcat::mc "Stop"] -width 6 -command {::AuPlayer::StopButton}
  button .$Local(playerframe).play -text [msgcat::mc "Play"] -foreground blue -activeforeground blue \
  -state disabled -command {::AuPlayer::PlayButton}
  
  # the placement 
  grid .$Local(playerframe).filedir_comment -row 0 -column 0 -sticky w 
  grid .$Local(playerframe).filedir -row 0 -column 2 -sticky ew
  
  ::FR_general::Line $Local(playerframe) 1 3 

  grid .$Local(playerframe).volume_comment -row 2 -column 0 -sticky w
  grid .$Local(playerframe).volume -row 2 -column 2 -sticky ew
  
  ::FR_general::Line $Local(playerframe) 3 3 
  
  grid .$Local(playerframe).filelist_selectall -row 4 -column 0 -sticky w
  grid .$Local(FileListFrame) -row 5 -column 0 -columnspan 3 -sticky nsew -pady 5 
  grid .$Local(playerframe).stop -row 6 -column 0 -sticky ew 
  grid .$Local(playerframe).play -row 6 -column 1 -columnspan 2 -sticky ew 
  
  grid columnconfigure .$Local(playerframe) 1 -weight 1
  grid rowconfigure .$Local(playerframe) 5 -weight 1 
 
  # the scroll window
  set Local(TitleFrame) [::FR_general::ScrollWindow $Local(FileListFrame)] 

  ::FR_balloonHelp::AddHelp ".$Local(playerframe).filedir" [msgcat::mc "This is the directory with your Sun audio files"]
  
  ::AuTags::BalloonHelpStatus
}


####################################################################
# load the directory with the au-files
####################################################################
proc ::AuPlayer::FileDir {} {
  variable Config; variable Local

  # output directory because we use the output of the tagger or cdripper sub-programs
  set OutputDir [::FR_dirBox::Create -initialdir [lindex $Config(OutputDir) 0] -parent .$Local(playerframe) \
  -title [msgcat::mc "Choose the directory with your au-files:"] -filetype [list [msgcat::mc "Sun audio files"] "*.au"]]
  
  # check if the answer is empty
  if {[lindex $OutputDir 0] != ""} {
	# copy the new directory (in list format)
    set Config(OutputDir) $OutputDir
    
    cd [lindex $Config(OutputDir) 0]

    ::FR_msgWindow::Add -format Info \
	-msg [msgcat::mc "The directory was set to: %s" [lindex $Config(OutputDir) 0]]
    
    ::AuPlayer::TitleFieldsCreate
  }
}


####################################################################
# select or deselect all songs in the title list
####################################################################
proc ::AuPlayer::SelectAllSongs {} {
  variable Local; 

  # we toggle the flag first
  if {$Local(SelectAllSongs)} {
	set Local(SelectAllSongs) 0
	# and adapt the text of the button
	.$Local(playerframe).filelist_selectall configure -text [msgcat::mc "Select all songs"]
  } else {
	set Local(SelectAllSongs) 1
	# and adapt the text of the button
	.$Local(playerframe).filelist_selectall configure -text [msgcat::mc "Deselect all songs"]
  } 
  
  for {set Idx 0} {$Idx < $Local(FileListItems)} {incr Idx} {
   # select or deselect all tracks
   set Local(SelectSong$Idx) $Local(SelectAllSongs)
  } 
}


####################################################################
# the play button
####################################################################
proc ::AuPlayer::PlayButton {{StartIdx "-1"}} {
  variable Local
  
  # we set the current played song below the list to can find
  # song 1 if selected
  set Local(CurrentPlayedSong) $StartIdx 
  
  # in case of an error we stop the loop over all selected songs
  set RunLoop 1
  
  # we try to play all selected songs
  while { $RunLoop } {
    set NextSong [::AuPlayer::FindNextSong $Local(CurrentPlayedSong)]

	# we separate the index from the return list
    set Idx [lindex $NextSong 0] 
  
	if {$Idx == -1} {
	  # no further songs found we stop the while loop
	  break
	} 
	  
	# we found a next song
	set Local(CurrentPlayedSong) $Idx 
    
	# start blinking of the active checkbutton
	::AuPlayer::StartBlinking
	
	# start playing
	if {[::AuPlayer::Play [lindex $NextSong 1]]} {
	  # we can not play the file, we stop the while loop also
	  set RunLoop 0
    }

    # stop blinking of the active checkbutton
	::AuPlayer::StopBlinking
	
	# check if the stop button was pressed
	if {$Local(Done) == 2} {
	  # we stop the playing of all further songs
	  set RunLoop 0
	}
  }
}


####################################################################
# the stop button
####################################################################
proc ::AuPlayer::StopButton {} {
  variable Local 
  
  # we set the finished flag in the play procedure
  set Local(Done) 2 
}


####################################################################
# Start blinking of the current selected song
####################################################################
proc ::AuPlayer::StartBlinking {} {
  variable Local 

  set Local(BlinkId) [after 2000 ::AuPlayer::AlternateBlinking]
}


####################################################################
# Alternate the checkbutton of the current selected song
####################################################################
proc ::AuPlayer::AlternateBlinking {} {
  variable Local

  # we alternate the setting
  if {$Local(SelectSong$Local(CurrentPlayedSong))} {
    set Local(SelectSong$Local(CurrentPlayedSong)) 0
  } else {
    set Local(SelectSong$Local(CurrentPlayedSong)) 1
  }

  # we call the procedure again in 2s
  set Local(BlinkId) [after 2000 ::AuPlayer::AlternateBlinking] 
}


####################################################################
# Start blinking of the current selected song
####################################################################
proc ::AuPlayer::StopBlinking {} {
  variable Local

  # we stop the next, already planed event
  after cancel $Local(BlinkId)

  # leave the function with a deselected checkbutton
  set Local(SelectSong$Local(CurrentPlayedSong)) 0
}


####################################################################
# Find the next selected song
####################################################################
proc ::AuPlayer::FindNextSong {StartIdx} {
  variable Local
  
  set Found 0 
  
  # the start index is the currently played song
  set Idx [expr {$StartIdx + 1}]

  while {$Idx < $Local(FileListItems)} {
    if {$Local(SelectSong$Idx)} {
      set Found 1
	  break
	}
    incr Idx
  } 
  
  if {$Found} {
	# we return the found index (for the next search) and the file name of the song
	return [list $Idx [lindex $Local(AllFiles) $Idx]]
  } else {
	# we didn't find anything, we return minus one
	return "-1"
  }
} 


####################################################################
# Find the next selected song
####################################################################
proc ::AuPlayer::PlayChanReadable {} {
  variable Local 

  if {$Local(OS) == "NetBSD"} {
	set Output [gets $Local(Fid)]
	# audioplay complains a wrong file size which we can ignore
	if {[string first $Output "bogus"] == -1} {
      # this is not a wrong file size, in this case it is an error
	  # or the song is finished
	  set Local(Done) 1
	}   
  } else {
	# for all other OS
    set Local(Done) 1
  } 
}


####################################################################
# play the handed over song
####################################################################
proc ::AuPlayer::Play {FileName} {
  variable Local 

  # define the tool for playing
  switch -- $Local(OS) {
    "OpenBSD_Old" {
      set Cmd "sox $FileName -t ossdsp /dev/sound" 
    }
    "OpenBSD" {
      set Cmd "aucat -i $FileName"
    }
    "NetBSD" {
      set Cmd "audioplay $FileName"
    }
    "Linux" {
	    # aplay doesn't detect au file format automatically,
	    # quiet mode to suppress the output of the sound format
      set Cmd "aplay -q -f cdr $FileName"
    }
  }

  # we use this variable to detect the end of the song
  set Local(Done) 0
  
  # we direct the error messages also to the standard output
  # (the channel becomes so readable in case of success or error)
  if {[catch { open "|$Cmd 2>@ stdout" r} Local(Fid)]} {
    ::FR_msgWindow::Add -format Error -msg [msgcat::mc "We could not start playing song: %s" $FileName] 
    return 1
  } 
  
  # save the program id for the stop function
  set Pid [pid $Local(Fid)]
  
  # set the channel into the non-blocking mode (for the stop button)
  fconfigure $Local(Fid) -blocking 0
  # the channel is readable in case of an error or if it is finished
  fileevent $Local(Fid) readable {::AuPlayer::PlayChanReadable}
  
  ::FR_msgWindow::Add -format Info -msg [msgcat::mc "We are currently playing: %s" $FileName] 
  	
  # we wait until the variable is set
  tkwait variable ::AuPlayer::Local(Done)
  
  # close the channel
  fileevent $Local(Fid) readable {}
  catch {close $Local(Fid)} 
  
  # in case of the stop function we have to kill the background task also
  if { $Local(Done) == 2 } {
    catch {exec kill $Pid} 
  } 
  
  return 0 
}


####################################################################
# generate one line per song and fill the entry-widgets with the 
# title/artist information
####################################################################
proc ::AuPlayer::TitleFieldsCreate {} {
  variable Config; variable Local
  
  # we destroy all existing widgets in the scroll frame
  if { $Local(FileListItems) != 0 } {
    foreach Widget [winfo children .$Local(TitleFrame)] {
      catch {destroy $Widget}
    }
    # reset the count of title fields
    set Local(FileListItems) 0
  }
  
  set Local(AllFiles) [::AuTags::CheckDirContent [lindex $Config(OutputDir) 0] "SunAudio"]

  # check if we found something
  if {$Local(AllFiles) == ""} {
	# deactivate the play button
    .$Local(playerframe).play configure -state disabled
    .$Local(playerframe).filelist_selectall configure -state disabled
    
	::FR_msgWindow::Add -format Error -msg [msgcat::mc "I didn't find any au-files. Have you chosen the right directory?"]
    # we can stop here
	return
  } 
   
  # the labels in the head of the table
  label .$Local(TitleFrame).file_comment -text [msgcat::mc "Filename"]
  label .$Local(TitleFrame).artist_comment -text [msgcat::mc "Artist"]
  label .$Local(TitleFrame).title_comment -text [msgcat::mc "Title"]

  # we place the header
  grid .$Local(TitleFrame).file_comment -row 0 -column 1 -sticky w
  grid .$Local(TitleFrame).artist_comment -row 0 -column 2 -sticky w 
  grid .$Local(TitleFrame).title_comment -row 0 -column 3 -sticky w
	
  # the start index
  set Index 0
    
  # now we process each file
  foreach File $Local(AllFiles) { 
    # we limit the count of characters in the file name to 20
    if { [string length $File] > 20 } {
	  # split file name into name and extension
      set RootName [file rootname $File]
      set FileExtension [file extension $File] 
      # range from 0...18 are 19 characters + "~" = 20
      # we reduce the file name length by the extension length
	  set NameLimit [expr {18 - [string length $FileExtension]}]
      # we add all together again
	  set FileName [string range $RootName 0 $NameLimit]
      append FileName "~" $FileExtension
    } else {
	  # the standard case: we can show the full file name
	  set FileName $File
    }
   
    # the file depending widgets
	set Local(SelectSong$Index) 0
	checkbutton .$Local(TitleFrame).select$Index -variable ::AuPlayer::Local(SelectSong$Index)
	label .$Local(TitleFrame).file$Index -text "$FileName"
    entry .$Local(TitleFrame).artist$Index -state normal -width 28 
	entry .$Local(TitleFrame).title$Index -state normal -width 28

	# the first index is zero but the first row is the header, so the row is index plus one
	set Row [expr {$Index + 1}] 
	grid .$Local(TitleFrame).select$Index -row $Row -column 0 -sticky w
	grid .$Local(TitleFrame).file$Index -row $Row -column 1 -sticky w
	grid .$Local(TitleFrame).artist$Index -row $Row -column 2 -sticky ew
	grid .$Local(TitleFrame).title$Index -row $Row -column 3 -sticky ew
	 
    # now we read the tags from the au-files
	set TagList [::AuFile::ReadTags [file join [lindex $Config(OutputDir) 0] $File]]
     
    # we check the success
	if { [lindex $TagList 0] != 0 } {
      # oops an error occurred - we ask for the reason
	  set ErrorList [::AuFile::GetErrorMsg]
    
	  ::FR_msgWindow::Add -format Error -msg [msgcat::mc "An error occurred during reading of file: %s" $File]
      ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "The reason: %s" [lindex $ErrorList 0]] 
      ::FR_msgWindow::Add -format BodyRed -msg [msgcat::mc "Details: %s" [lindex $ErrorList 1]]

	  # we need the count of widgets to destroy all at the next start of titleartist,
      # Index is zero based so we use Row because it is Index + 1
	  set Local(FileListItems) $Row
    
      return
    }

    # we generate lists with the information of all au-files
	# the form: success Title Artist Album Comment Year Track Genre (in form of text) 
    .$Local(TitleFrame).title$Index insert end [lindex $TagList 1]
    .$Local(TitleFrame).artist$Index insert end [lindex $TagList 2] 

    # we prepare the next index
	incr Index 
  }

  grid columnconfigure .$Local(TitleFrame) 2 -weight 1
  grid columnconfigure .$Local(TitleFrame) 3 -weight 1

  # we need the count of widgets to destroy all at the next start of titleartist,
  # we use Index here because it was already increased in the last loop 
  set Local(FileListItems) $Index 
  
  ::FR_general::ScrollWindowRefresh $Local(FileListFrame) 
  
  # we configure the scroll increment with help of the first entry-box
  ::FR_general::ScrollWindowIncrement $Local(FileListFrame) $Local(TitleFrame).artist0 
  
  .$Local(playerframe).play configure -state normal
  .$Local(playerframe).filelist_selectall configure -state normal

  # we select all songs as default
  ::AuPlayer::SelectAllSongs
} 


####################################################################
# set the volume level
####################################################################
proc ::AuPlayer::SetVolume { Level } {
  variable Local

  # start the OS depending mixer
  switch -- $Local(OS) {
    "OpenBSD" {
      ::AuPlayer::SetVolumeOpenBSD $Level
    }
    "OpenBSD_Old" -
    "NetBSD" {
      # elder OpenBSD and NetBSD use mixerctl
      SetVolumeNetBSD $Level
    }
    "Linux" {
      SetVolumeLinux $Level
    }
  }
}


####################################################################
# set the volume level in NetBSD and elder OpenBSD
####################################################################
proc ::AuPlayer::SetVolumeNetBSD { Level } {

  # transform the 0...100 scale to 0...255 used in mixerctl
  # it should be an integer calculation
  set NewVolume [expr {($Level * 255) / 100}] 
  
  # we create the mixerctl command with the new volume
  set Cmd "mixerctl -w outputs.master=${NewVolume},${NewVolume}" 
  
  # we execute the command
  if { [catch {eval exec $Cmd} ExecOutput] } {
	# command execution failed we throw an error
	::FR_msgWindow::Add -format Error -msg [msgcat::mc "Could not set the mixer level to: %s" $Level] 
  }
}


####################################################################
# set the volume level in Linux
####################################################################
proc ::AuPlayer::SetVolumeLinux { Level } {

  # we create the amixer command with the new volume
  set Cmd "amixer sset Master playback ${Level}%" 
  
  # we execute the command
  if { [catch {eval exec $Cmd} ExecOutput] } {
	# command execution failed we throw an error
	::FR_msgWindow::Add -format Error -msg [msgcat::mc "Could not set the mixer level to: %s" $Level] 
  }
}


####################################################################
# set the volume level in newer OpenBSD
####################################################################
proc ::AuPlayer::SetVolumeOpenBSD { Level } {

  # transform the 0...100 scale to 0.0 ... 1.0 used in sndioctl
  set NewVolume [expr {double($Level) / 100.0}] 
  
  # we create the sndioctl command with the new volume,
  # we set both (stereo) channels
  set Cmd "sndioctl output.level=${NewVolume}" 
  
  # we execute the command
  if { [catch {eval exec $Cmd} ExecOutput] } {
	# command execution failed we throw an error
	::FR_msgWindow::Add -format Error -msg [msgcat::mc "Could not set the mixer level to: %s" $Level] 
  }
}
